@@ -33,15 +33,16 @@
 
 ///---Internal Constants---///
 ///---Globals---///
+static xcb_generic_event_t *ev  = NULL;
 static void (*events[XCB_NO_OPERATION])(xcb_generic_event_t *e);
 static unsigned int numlockmask = 0;
-int sigcode;                        // Signal code. Non-zero if we've been interruped by a signal.
-xcb_connection_t *conn;             // Connection to X server.
-xcb_ewmh_connection_t *ewmh;        // Ewmh Connection.
-xcb_screen_t     *screen;           // Our current screen.
-int randrbase;                      // Beginning of RANDR extension events.
+int sigcode = 0;                        // Signal code. Non-zero if we've been interruped by a signal.
+xcb_connection_t *conn = NULL;             // Connection to X server.
+xcb_ewmh_connection_t *ewmh = NULL;        // Ewmh Connection.
+xcb_screen_t     *screen = NULL;           // Our current screen.
+int randrbase = 0;                      // Beginning of RANDR extension events.
 static uint8_t curws = 0;                  // Current workspace.
-struct client *focuswin;            // Current focus window.
+struct client *focuswin = NULL;            // Current focus window.
 static xcb_drawable_t top_win=0;           // Window always on top.
 static struct item *winlist = NULL;        // Global list of all client windows.
 static struct item *monlist = NULL;        // List of all physical monitor outputs.
@@ -56,6 +57,7 @@ xcb_atom_t ATOM[NB_ATOMS];
 ///---Functions prototypes---///
 static void run(void);
 static bool setup(int);
+static void install_sig_handlers(void);
 static void start(const Arg *);
 static void mousemotion(const Arg *);
 static void cursor_move(const Arg *);
@@ -76,6 +78,7 @@ static void changescreen(const Arg *);
 static void grabkeys(void);
 static void twobwm_restart();
 static void twobwm_exit();
+static void centerpointer(xcb_drawable_t, struct client *);
 static void always_on_top();
 static bool setup_keyboard(void);
 static bool setupscreen(void);
@@ -88,6 +91,9 @@ static void raise_current_window(void);
 static void raiseorlower();
 static void setunfocus(void);
 static void maximize(const Arg *);
+static void fullscreen(const Arg *);
+static void unmaxwin(struct client *);
+static void maxwin(struct client *, uint8_t);
 static void maximize_helper(struct client *,uint16_t, uint16_t, uint16_t, uint16_t);
 static void hide();
 static void clientmessage(xcb_generic_event_t *);
@@ -125,7 +131,7 @@ static struct monitor *findmonitor(xcb_randr_output_t);
 static struct monitor *findclones(xcb_randr_output_t, const int16_t, const int16_t);
 static struct monitor *findmonbycoord(const int16_t, const int16_t);
 static void delmonitor(struct monitor *);
-static struct monitor *addmonitor(xcb_randr_output_t, char *,const int16_t, const int16_t, const uint16_t, const uint16_t);
+static struct monitor *addmonitor(xcb_randr_output_t, const int16_t, const int16_t, const uint16_t, const uint16_t);
 static void raisewindow(xcb_drawable_t);
 static void movelim(struct client *);
 static void movewindow(xcb_drawable_t, const int16_t, const int16_t);
@@ -139,7 +145,7 @@ static void mouseresize(struct client *,const int16_t,const int16_t);
 static void setborders(struct client *,const bool);
 static void unmax(struct client *);
 static bool getpointer(const xcb_drawable_t *, int16_t *,int16_t *);
-static bool getgeom(const xcb_drawable_t *, int16_t *, int16_t *, uint16_t *,uint16_t *);
+static bool getgeom(const xcb_drawable_t *, int16_t *, int16_t *, uint16_t *,uint16_t *, uint8_t *);
 static void configwin(xcb_window_t, uint16_t,const struct winconf *);
 static void sigcatch(const int);
 static void ewmh_init(void);
@@ -213,12 +219,6 @@ twobwm_exit()
 	exit(EXIT_SUCCESS);
 }
 
-void
-sigcatch(const int sig)
-{
-	sigcode = sig;
-}
-
 void
 saveorigsize(struct client *client)
 {
@@ -313,15 +313,33 @@ movepointerback(const int16_t startx, const int16_t starty,
 void
 cleanup(void)
 {
+    free(ev);
+    if(monlist)
+        delallitems(&monlist,NULL);
+    struct item *curr,*wsitem;
+    for(int i = 0; i < WORKSPACES; i++){
+        if(!wslist[i])
+            continue;
+        curr = wslist[i];
+        while(curr){
+            wsitem = curr;
+            curr = curr->next;
+            free(wsitem);
+        }
+    }
+    if(winlist){
+        delallitems(&winlist,NULL);
+    }
+    if (ewmh != NULL){
+	    xcb_ewmh_connection_wipe(ewmh);
+		free(ewmh);
+    }
+    if(!conn){
+        return;
+    }
 	xcb_set_input_focus(conn, XCB_NONE,XCB_INPUT_FOCUS_POINTER_ROOT,
 			XCB_CURRENT_TIME);
-	delallitems(wslist, NULL);
-	xcb_ewmh_connection_wipe(ewmh);
 	xcb_flush(conn);
-
-	if (ewmh != NULL)
-		free(ewmh);
-
 	xcb_disconnect(conn);
 }
 
@@ -890,21 +908,21 @@ setupwin(xcb_window_t win)
 	xcb_ewmh_get_atoms_reply_t win_type;
 	struct item *item;
 	struct client *client;
-
 	if (xcb_ewmh_get_wm_window_type_reply(ewmh,
-				xcb_ewmh_get_wm_window_type(ewmh, win),
-				&win_type, NULL) == 1)
+                xcb_ewmh_get_wm_window_type(ewmh, win), &win_type, NULL)
+                == 1)
+    {
 		for (i = 0; i < win_type.atoms_len; i++) {
 			a = win_type.atoms[i];
 			if (a == ewmh->_NET_WM_WINDOW_TYPE_TOOLBAR || a
 					== ewmh->_NET_WM_WINDOW_TYPE_DOCK || a
 					== ewmh->_NET_WM_WINDOW_TYPE_DESKTOP ) {
-				xcb_ewmh_get_atoms_reply_wipe(&win_type);
 				xcb_map_window(conn,win);
 				return NULL;
 			}
 		}
-
+        xcb_ewmh_get_atoms_reply_wipe(&win_type);
+    }
 	values[0] = XCB_EVENT_MASK_ENTER_WINDOW;
 	xcb_change_window_attributes(conn, win, XCB_CW_BACK_PIXEL,
 			&conf.empty_col);
@@ -947,7 +965,7 @@ setupwin(xcb_window_t win)
 
 	/* Get window geometry. */
 	getgeom(&client->id, &client->x, &client->y, &client->width,
-			&client->height);
+			&client->height, &client->depth);
 
 	/* Get the window's incremental size step, if any.*/
 	xcb_icccm_get_wm_normal_hints_reply(conn,
@@ -1049,7 +1067,7 @@ setup_keyboard(void)
 
 	numlock = xcb_get_keycodes(XK_Num_Lock);
 
-	for (i=0; i<8; i++) {
+	for (i=4; i<8; i++) {
 		for (j=0; j<reply->keycodes_per_modifier; j++) {
 			xcb_keycode_t keycode = modmap[i
 				* reply->keycodes_per_modifier + j];
@@ -1236,11 +1254,10 @@ getoutputs(xcb_randr_output_t *outputs, const int len,
 				NULL)) == NULL)
 			continue;
 
-	name_len = MIN(16, xcb_randr_get_output_info_name_length(output));
-	name = malloc(name_len+1);
-
-	snprintf(name, name_len+1, "%.*s", name_len,
-			xcb_randr_get_output_info_name(output));
+	//name_len = MIN(16, xcb_randr_get_output_info_name_length(output));
+	//name = malloc(name_len+1);
+	//snprintf(name, name_len+1, "%.*s", name_len,
+	//		xcb_randr_get_output_info_name(output));
 
 	if (XCB_NONE != output->crtc) {
 		icookie = xcb_randr_get_crtc_info(conn, output->crtc,
@@ -1258,7 +1275,7 @@ getoutputs(xcb_randr_output_t *outputs, const int len,
 
 		/* Do we know this monitor already? */
 		if (NULL == (mon = findmonitor(outputs[i])))
-			addmonitor(outputs[i], name, crtc->x, crtc->y,
+			addmonitor(outputs[i], crtc->x, crtc->y,
 					crtc->width,crtc->height);
 		else
 			/* We know this monitor. Update information.
@@ -1376,7 +1393,7 @@ findmonbycoord(const int16_t x, const int16_t y)
 }
 
 struct monitor *
-addmonitor(xcb_randr_output_t id, char *name,const int16_t x, const int16_t y,
+addmonitor(xcb_randr_output_t id, const int16_t x, const int16_t y,
 		const uint16_t width,const uint16_t height)
 {
 	struct item *item;
@@ -1390,7 +1407,6 @@ addmonitor(xcb_randr_output_t id, char *name,const int16_t x, const int16_t y,
 
 	item->data  = mon;
 	mon->id     = id;
-	mon->name   = name;
 	mon->item   = item;
 	mon->x      = x;
 	mon->y      = y;
@@ -1440,9 +1456,8 @@ movelim(struct client *client)
 	noborder(&temp, client, true);
 
 	/* Is it outside the physical monitor or close to the side? */
-	if (client->y-conf.borderwidth < mon_y)
-		client->y = mon_y;
-	else if (client->y < borders[2] + mon_y)
+	if (client->y-conf.borderwidth < mon_y
+		|| client->y < borders[2] + mon_y)
 		client->y = mon_y;
 	else if (client->y + client->height + (conf.borderwidth * 2) > mon_y
 			+ mon_height - borders[2])
@@ -1456,11 +1471,6 @@ movelim(struct client *client)
 		client->x = mon_x + mon_width - client->width
 			- conf.borderwidth * 2;
 
-	if (client->y + client->height > mon_y + mon_height
-			- conf.borderwidth * 2)
-		client->y = (mon_y + mon_height - conf.borderwidth * 2)
-			- client->height;
-
 	movewindow(client->id, client->x, client->y);
 	noborder(&temp, client, false);
 }
@@ -1641,8 +1651,8 @@ start(const Arg *arg)
 	if (fork())
 		return;
 
-	if (conn)
-		close(screen->root);
+//	if (conn)
+//		close(screen->root);
 
 	setsid();
 	execvp((char*)arg->com[0], (char**)arg->com);
@@ -1979,9 +1989,7 @@ setborders(struct client *client,const bool isitfocused)
 	};
 
 	xcb_pixmap_t pmap = xcb_generate_id(conn);
-	/* my test have shown that drawing the pixmap directly on the root
-	 * window is faster then drawing it on the window directly */
-	xcb_create_pixmap(conn, screen->root_depth, pmap, screen->root,
+	xcb_create_pixmap(conn, client->depth, pmap, client->id,
 			client->width + (conf.borderwidth * 2),
 			client->height + (conf.borderwidth * 2)
 	);
@@ -2061,24 +2069,47 @@ unmax(struct client *client)
 void
 maximize(const Arg *arg)
 {
+	maxwin(focuswin, 1);
+}
+
+void
+fullscreen(const Arg *arg)
+{
+	maxwin(focuswin, 0);
+}
+
+
+void
+unmaxwin(struct client *client){
+	unmax(client);
+	client->maxed = false;
+	setborders(client,true);
+	xcb_change_property(conn, XCB_PROP_MODE_REPLACE, client->id,
+			ewmh->_NET_WM_STATE, XCB_ATOM_ATOM, 32, 0, NULL);
+}
+
+void 
+maxwin(struct client *client, uint8_t with_offsets){
 	uint32_t values[4];
 	int16_t mon_x, mon_y;
-	uint16_t mon_width, mon_height;
+	int16_t mon_width, mon_height;
 
 	if (NULL == focuswin)
 		return;
 
 	/* Check if maximized already. If so, revert to stored geometry. */
 	if (focuswin->maxed) {
-		unmax(focuswin);
-		focuswin->maxed = false;
-		setborders(focuswin,true);
+		unmaxwin(focuswin);
 		return;
 	}
 
-	getmonsize(arg->i, &mon_x, &mon_y, &mon_width, &mon_height, focuswin);
-	maximize_helper(focuswin, mon_x, mon_y, mon_width, mon_height);
+	getmonsize(with_offsets, &mon_x, &mon_y, &mon_width, &mon_height, client);
+	maximize_helper(client, mon_x, mon_y, mon_width, mon_height);
 	raise_current_window();
+	if (!with_offsets) {
+		xcb_change_property(conn, XCB_PROP_MODE_REPLACE, client->id,
+			ewmh->_NET_WM_STATE, XCB_ATOM_ATOM, 32, 1, &ewmh->_NET_WM_STATE_FULLSCREEN);
+	}
 	xcb_flush(conn);
 }
 
@@ -2250,7 +2281,7 @@ getpointer(const xcb_drawable_t *win, int16_t *x, int16_t *y)
 
 bool
 getgeom(const xcb_drawable_t *win, int16_t *x, int16_t *y, uint16_t *width,
-		uint16_t *height)
+		uint16_t *height, uint8_t *depth)
 {
 	xcb_get_geometry_reply_t *geom = xcb_get_geometry_reply(conn,
 			xcb_get_geometry(conn, *win), NULL);
@@ -2262,6 +2293,7 @@ getgeom(const xcb_drawable_t *win, int16_t *x, int16_t *y, uint16_t *width,
 	*y = geom->y;
 	*width = geom->width;
 	*height = geom->height;
+	*depth = geom->depth;
 
 	free(geom);
 	return true;
@@ -2480,10 +2512,13 @@ handle_keypress(xcb_generic_event_t *e)
 	xcb_key_press_event_t *ev       = (xcb_key_press_event_t *)e;
 	xcb_keysym_t           keysym   = xcb_get_keysym(ev->detail);
 
-	for (unsigned int i=0; i<LENGTH(keys); i++)
+	for (unsigned int i=0; i<LENGTH(keys); i++) {
 		if (keysym == keys[i].keysym && CLEANMASK(keys[i].mod)
-				== CLEANMASK(ev->state) && keys[i].func)
+				== CLEANMASK(ev->state) && keys[i].func) {
 			keys[i].func(&keys[i].arg);
+			break;
+		}
+	}
 }
 
 /* Helper function to configure a window. */
@@ -2798,8 +2833,13 @@ buttonpress(xcb_generic_event_t *ev)
 				== CLEANMASK(e->state)){
 			if ((focuswin==NULL) && buttons[i].func == mousemotion)
 				return;
-
-			buttons[i].func(&(buttons[i].arg));
+			if (buttons[i].root_only) {
+				if (e->event == e->root && e->child == 0)
+					buttons[i].func(&(buttons[i].arg));
+			}
+			else {
+				buttons[i].func(&(buttons[i].arg));
+			}
 		}
 }
 
@@ -2832,6 +2872,31 @@ clientmessage(xcb_generic_event_t *ev)
 	}
 	else if (e->type == ewmh->_NET_CURRENT_DESKTOP)
 		changeworkspace_helper(e->data.data32[0]);
+	else if (e->type == ewmh->_NET_WM_STATE && e->format == 32) {
+		cl = findclient(&e->window);
+		if (NULL == cl)
+			return;
+		if(e->data.data32[1] == ewmh->_NET_WM_STATE_FULLSCREEN
+				|| e->data.data32[2] == ewmh->_NET_WM_STATE_FULLSCREEN) {
+			switch (e->data.data32[0]) {
+				case XCB_EWMH_WM_STATE_REMOVE:
+					unmaxwin(cl);
+					break;
+				case XCB_EWMH_WM_STATE_ADD:
+					maxwin(cl, false);
+					break;
+				case XCB_EWMH_WM_STATE_TOGGLE:
+						if(cl->maxed)
+							unmaxwin(cl);
+						else
+							maxwin(cl, false);
+					break;
+
+				default:
+					break;
+			}
+		}
+	}
 }
 
 void
@@ -2954,16 +3019,16 @@ confignotify(xcb_generic_event_t *ev)
 void
 run(void)
 {
-	xcb_generic_event_t *ev;
 	sigcode = 0;
 
 	while (0 == sigcode) {
 		/* the WM is running */
 		xcb_flush(conn);
 
-		if (xcb_connection_has_error(conn))
+		if (xcb_connection_has_error(conn)){
+            cleanup();
 			abort();
-
+        }
 		if ((ev = xcb_wait_for_event(conn))) {
 			if(ev->response_type==randrbase +
 					XCB_RANDR_SCREEN_CHANGE_NOTIFY)
@@ -3018,15 +3083,17 @@ grabbuttons(struct client *c)
 	};
 
 	for (unsigned int b=0; b<LENGTH(buttons); b++)
-		for (unsigned int m=0; m<LENGTH(modifiers); m++)
-			xcb_grab_button(conn, 1, c->id,
-					XCB_EVENT_MASK_BUTTON_PRESS,
-					XCB_GRAB_MODE_ASYNC,
-					XCB_GRAB_MODE_ASYNC,
-					screen->root, XCB_NONE,
-					buttons[b].button,
-					buttons[b].mask|modifiers[m]
-			);
+		if (!buttons[b].root_only) {
+			for (unsigned int m=0; m<LENGTH(modifiers); m++)
+				xcb_grab_button(conn, 1, c->id,
+						XCB_EVENT_MASK_BUTTON_PRESS,
+						XCB_GRAB_MODE_ASYNC,
+						XCB_GRAB_MODE_ASYNC,
+						screen->root, XCB_NONE,
+						buttons[b].button,
+						buttons[b].mask|modifiers[m]
+				);
+		}
 }
 
 void
@@ -3036,7 +3103,10 @@ ewmh_init(void)
 		printf("Fail\n");
 
 	xcb_intern_atom_cookie_t *cookie = xcb_ewmh_init_atoms(conn, ewmh);
-	xcb_ewmh_init_atoms_replies(ewmh, cookie, (void *)0);
+	if(!xcb_ewmh_init_atoms_replies(ewmh, cookie, (void *)0)){
+        fprintf(stderr,"%s\n","xcb_ewmh_init_atoms_replies:faild.");
+        exit(1);
+    }
 }
 
 bool
@@ -3057,20 +3127,9 @@ setup(int scrno)
 	if (!screen)
 		return false;
 
-	xcb_window_t cwin = xcb_generate_id(conn);
-
-	xcb_create_window(conn, XCB_COPY_FROM_PARENT, cwin, screen->root, 0, 0,
-			screen->width_in_pixels, screen->height_in_pixels, 0,
-			XCB_WINDOW_CLASS_INPUT_OUTPUT,
-			XCB_COPY_FROM_PARENT,XCB_CW_EVENT_MASK,
-			event_mask_pointer
-	);
-
 	ewmh_init();
-	xcb_ewmh_set_wm_pid(ewmh, cwin, getpid());
-	xcb_ewmh_set_wm_name(ewmh, cwin, 4, "2bwm");
-	xcb_ewmh_set_supporting_wm_check(ewmh, cwin, cwin);
-	xcb_ewmh_set_supporting_wm_check(ewmh, screen->root, cwin);
+	xcb_ewmh_set_wm_pid(ewmh, screen->root, getpid());
+	xcb_ewmh_set_wm_name(ewmh, screen->root, 4, "2bwm");
 
 	xcb_atom_t net_atoms[] = {
 		ewmh->_NET_SUPPORTED,              ewmh->_NET_WM_DESKTOP,
@@ -3083,7 +3142,8 @@ setup(int scrno)
 		ewmh->_NET_WM_WINDOW_TYPE_TOOLBAR, ewmh->_NET_WM_PID,
 		ewmh->_NET_CLIENT_LIST,            ewmh->_NET_CLIENT_LIST_STACKING,
 		ewmh->WM_PROTOCOLS,                ewmh->_NET_WM_STATE,
-		ewmh->_NET_WM_STATE_DEMANDS_ATTENTION
+		ewmh->_NET_WM_STATE_DEMANDS_ATTENTION,
+		ewmh->_NET_WM_STATE_FULLSCREEN
 	};
 
 	xcb_ewmh_set_supported(ewmh, scrno, LENGTH(net_atoms), net_atoms);
@@ -3156,11 +3216,13 @@ setup(int scrno)
 				XCB_CW_EVENT_MASK, values));
 	xcb_flush(conn);
 
-	if (error)
+	if (error){
+        fprintf(stderr,"%s\n","xcb_request_check:faild.");
+        free(error);
 		return false;
-
-	xcb_ewmh_set_current_desktop(ewmh, 0, curws);
-	xcb_ewmh_set_number_of_desktops(ewmh, 0, WORKSPACES);
+    }
+	xcb_ewmh_set_current_desktop(ewmh, scrno, curws);
+	xcb_ewmh_set_number_of_desktops(ewmh, scrno, WORKSPACES);
 
 	grabkeys();
 	/* set events */
@@ -3195,12 +3257,15 @@ twobwm_restart(void)
 	execvp(TWOBWM_PATH, NULL);
 }
 
-int
-main(void)
+void
+sigcatch(const int sig)
 {
-	int scrno;
+	sigcode = sig;
+}
 
-	/* Install signal handlers. */
+void
+install_sig_handlers(void)
+{
 	struct sigaction sa;
 	struct sigaction sa_chld;
 	sa.sa_handler = SIG_IGN;
@@ -3211,20 +3276,22 @@ main(void)
 		exit(-1);
 	sa.sa_handler = sigcatch;
 	sigemptyset(&sa.sa_mask);
-	sa.sa_flags = SA_RESTART; /* Restart functions if
-								 interrupted by handler */
-	if (sigaction(SIGINT, &sa, NULL) == -1)
+	sa.sa_flags = SA_RESTART; /* Restart if interrupted by handler */
+	if ( sigaction(SIGINT, &sa, NULL) == -1
+		|| sigaction(SIGHUP, &sa, NULL) == -1
+		|| sigaction(SIGTERM, &sa, NULL) == -1)
 		exit(-1);
-	if (sigaction(SIGHUP, &sa, NULL) == -1)
-		exit(-1);
-	if (sigaction(SIGTERM, &sa, NULL) == -1)
-		exit(-1);
-	atexit(cleanup);
+}
 
+int
+main(int argc, char **argv)
+{
+	int scrno = 0;
+	atexit(cleanup);
+	install_sig_handlers();
 	if (!xcb_connection_has_error(conn = xcb_connect(NULL, &scrno)))
 		if (setup(scrno))
 			run();
-
 	/* the WM has stopped running, because sigcode is not 0 */
 	exit(sigcode);
 }
